from ..common import DtoObject


class ChampionDto(DtoObject):
    pass


class ChampionListDto(DtoObject):
    pass


class ChampionReleasesDto(DtoObject):
    pass


class ChampionReleaseDto(DtoObject):
    pass


class ChampionAllRatesDto(DtoObject):
    pass


class ChampionRatesDto(DtoObject):
    pass
